function [constraints bboxes]=inference(structure,sw,w,lambda,c_appr,c_proj,spry,attrIdx,integralImgDir,IMDB,imgDir,printDir)
% this function infers one patch per image according to the structure,
% which minimizes the overall objective function 
% NOTE: the stack variable holds the index and map of all nodes need to be
% visited, which means that all map are hold in memory, which further
% implies that this implementation will work for chain but may need some
% modifications for tree (e.g. to save map to disk)

structData=structure.data;
nodeIdx=cellfun(@(x)x.nodeIdx,structData);
fprintf('Inference launched: Node#=%g, c_appr=%g, c_proj=%g, lambda=%g\n',length(structData),c_appr,c_proj,lambda);

for i = 1:length(structData)
    structData{i}.sw=sw;
end

% Walk from leaves to root of tree, passing message to parent
for k = length(structData):-1:2,
    par=find(structData{k}.parent==nodeIdx);
    tic
    [structData{k} structData{par}]=passmsg(structData{k},structData{par},w,lambda,c_appr,c_proj,spry,integralImgDir);
    toc
    
    % backtrack now
    [res ptr score imgIdx]=backtrack(structData(k-1:end));
    
    % printout
    if ~isempty(res)
        printImageBbox(imgIdx,res,score,imgDir,IMDB,fullfile(printDir,'backtrack'),size(res{1},1),struct('lambda',lambda,'c_appr',c_appr,'c_proj',c_proj));
    end
    fprintf('Done inference from image %g --> image %g\n',k,k-1);
    
end
[constraints bboxes]=organizeOutput(structData,imgIdx,res,score,attrIdx);

function [constraints bboxes]=organizeOutput(structData,imgIdx,res,score,attrIdx)
% forming output for next iter ranker learning
nodeIdx=cellfun(@(x)x.nodeIdx,structData);
[maxScore maxIdx]=max(score);
res=res{maxIdx};
bboxes=zeros(size(res,1),7);
for i = 1:size(res,1)
    bboxes(i,:)=[res(i,1:4) imgIdx(i) res(i,8)-res(i,6)+1 res(i,7)-res(i,5)+1];
end

[sortVal sortIdx]=sort(res(:,11),'descend');
sortedNodeIdx=res(sortIdx,13);
[ignore corrIdx]=ismember(sortedNodeIdx,nodeIdx);
structData=structData(corrIdx);
constraints=[];
for i = 1:length(structData)-1
    constraints=[constraints; structData{i}.imgIdx structData{i+1}.imgIdx attrIdx 1 1 1 1]; % we set the confidence of all relative attributes to be uniform
end

function printImageBbox(imgIdx,res,score,imgDir,IMDB,printDir,idx,params)
% printout: NOTE that currently assumes all res share the same imgIdx
topK=10;% display topK tracks
[sortedScore sortIdx]=sort(score,'descend');
res=res(sortIdx(1:min(length(sortIdx),topK)));
score=score(sortIdx(1:min(length(sortIdx),topK)));

% write images
for i = 1:length(imgIdx)
    im=readImIMDB(imgIdx(i),IMDB);
    bboxes=cellfun(@(x)x(i,:),res,'uniformoutput',false);
    bboxes=cat(1,bboxes{:});
    bboxes(:,1:4)=clip_to_image(round(bboxes(:,1:4)),[1 1 size(im,2) size(im,1)]);
    for j = 1:size(bboxes,1)
        filename=fullfile(printDir,sprintf('%g',idx),sprintf('%g',j),sprintf('%g.jpg',imgIdx(i)));
        plottedIm=plot_bb(im,bboxes(j,:),'r');
        myimwrite(plottedIm,filename);
    end
end

% make html
htmlFilename=fullfile(printDir,sprintf('%g.html',idx));
N=zeros(length(res),1);
imgList=[];
imgStr=[];
for i = 1:length(res)
    N(i)=size(res{i},1);
    for j = 1:N(i)
        imgList{end+1}=relativepath(fullfile(printDir,sprintf('%g',idx),sprintf('%g',i),sprintf('%g.jpg',imgIdx(j))),printDir);
        if j==1
            imgStr{end+1}=sprintf('<font color="red">(%g/%g) lambda=%g,c_appr=%g,c_proj=%g</font> ||  score=%.3f,img:%g,r=%.3e',...
                i,length(res),params.lambda,params.c_appr,params.c_proj,score(i),imgIdx(j),res{i}(j,11));
        else
            imgStr{end+1}=sprintf('img:%g,r=%.3e,d2p=%g',imgIdx(j),res{i}(j,11),res{i}(j,12));
        end
    end
end
displayPatch(htmlFilename,imgList,imgStr,relativepath(fullfile(printDir,sprintf('%g.html',idx-1)),printDir),relativepath(fullfile(printDir,sprintf('%g.html',idx+1)),printDir),N,200,200);



function [res ptr score imgIdx]=backtrack(structData)
% format of res
% [1-4 bbox, 5-8 bbox on feat, 9 layer index, 10 layer scale, 11 resp, 12 appearance dist to parent, 13 nodeIdx]


map=structData{1}.score;
imgIdx=cellfun(@(x)x.imgIdx,structData);
h=fspecial('gaussian',4,4);
thresh=-inf;
pathStart=[];
score=[];
for i = 1:length(map)
    for j =1:length(map{i})
        [hptr,wptr]=find(imregionalmax(imfilter(map{i}{j},h,'same'),8) & (map{i}{j}>=thresh));
        if isempty(hptr)
            continue;
        end
        pathStart=[pathStart;hptr wptr i*ones(length(hptr),1) j*ones(length(hptr),1)];
        idx=sub2ind(size(map{i}{j}),hptr,wptr);
        score=[score;map{i}{j}(idx)];
    end
end

res=cell(size(pathStart,1),1);
ptr=cell(size(pathStart,1),1);
for i = 1:size(pathStart,1)
    [res{i} ptr{i}]=backtrackRunner(pathStart(i,:),structData);
end

function [res ptr]=backtrackRunner(pathStart,structData)
% the function traces back from root to leave
res=zeros(length(structData),13);
ptr=nan(length(structData),5);
for i = 1:length(structData)
    if i==1
        info=getCoord(pathStart,structData{i}.sw,structData{i}.pyramidSize,structData{i}.stepsize,structData{i}.scales);
        ptr(i,:)=[pathStart structData{i}.nodeIdx];
        res(i,:)=[info structData{i}.resp{ptr(i,3)}{ptr(i,4)}(ptr(i,1),ptr(i,2)) nan structData{i}.nodeIdx];
    else
        par=find(structData{i}.parent==ptr(:,5));     
        hptr=structData{i}.lh{ptr(par,3)}{ptr(par,4)}(ptr(par,1),ptr(par,2));
        wptr=structData{i}.lw{ptr(par,3)}{ptr(par,4)}(ptr(par,1),ptr(par,2));
        layerptr=structData{i}.lk{ptr(par,3)}{ptr(par,4)}(ptr(par,1),ptr(par,2));
        apprDist=structData{i}.lad{ptr(par,3)}{ptr(par,4)}(ptr(par,1),ptr(par,2));
        [v u]=ind2sub([size(structData{i}.pyramidSize,1) size(structData{i}.sw,1)],layerptr);
        info=getCoord([hptr wptr u v],structData{i}.sw,structData{i}.pyramidSize,structData{i}.stepsize,structData{i}.scales);
        ptr(i,:)=[hptr wptr u v structData{i}.nodeIdx];
        res(i,:)=[info structData{i}.resp{ptr(i,3)}{ptr(i,4)}(ptr(i,1),ptr(i,2)) apprDist structData{i}.nodeIdx];
    end
end




function info=getCoord(ptr,sw,pyramidSize,stepsize,scales)
u=ptr(3);
v=ptr(4);
sw=sw(u,:);
pyramidSize=pyramidSize(v,:);
stepsize=stepsize(v);
[ignore hmark wmark]=genSize(pyramidSize,sw,stepsize);
hptr=hmark(ptr(1));
wptr=wmark(ptr(2));
pyramidBbox=[wptr hptr wptr+sw(2)-1 hptr+sw(1)-1 v scales(v)];
info=[pyramidBbox(1:4)/scales(v) pyramidBbox];



function [feat scales scaledSize]=loadIntegralMap(integralDir,imgIdx)
integralFeat=load(fullfile(integralDir,sprintf('%g.mat',imgIdx)));
scaledSize=integralFeat.scaledSize;
scales=integralFeat.scales;
feat=integralFeat.feat;
% TODO: remove the following
scaledSize=scaledSize(1:2:8,:);
scales=scales(1:2:8);
feat=feat(1:2:8);

function [contFeat M N hMark wMark atomN]=getBatchFeat(spry,intMap,SW,stepsize)
% assume that spry starts with "1" which means that the first atomN column
% of contFeat must be the feat of the global patch
M=size(intMap,1);
N=size(intMap,2);
ppp=reshape(1:M*N,[M N]);
intMap=reshape(intMap,[],size(intMap,3));
[ignore hMark wMark]=genSize([M N],SW,stepsize);
indMark=ppp(hMark,wMark);
indMark=indMark(:);
[uu vv]=ind2sub([M N],indMark);
atomN=size(intMap,2);
contFeat=zeros(length(indMark),atomN*sum(spry.^2));
counter=0;
for i = 1:length(spry)
    hGrid=round(SW(1)/spry(i));
    wGrid=round(SW(2)/spry(i));
    for j = 1:spry(i)^2
        [hIdx wIdx]=ind2sub([spry(i) spry(i)],j);
        uuStart=uu+(hIdx-1)*hGrid;
        vvStart=vv+(wIdx-1)*wGrid;
        uuEnd=uuStart+hGrid-1;
        vvEnd=vvStart+wGrid-1;
        bbox=clip_to_image([vvStart uuStart vvEnd uuEnd],[1 1 N M]);
        vvStart=bbox(:,1);
        uuStart=bbox(:,2);
        vvEnd=bbox(:,3);
        uuEnd=bbox(:,4);
        hSwS=sub2ind([M N],uuStart,vvStart);
        hEwE=sub2ind([M N],uuEnd,vvEnd);
        hSwE=sub2ind([M N],uuStart,vvEnd);
        hEwS=sub2ind([M N],uuEnd,vvStart);
        feat=intMap(hEwE,:)+intMap(hSwS,:)-intMap(hSwE,:)-intMap(hEwS,:);
        feat=bsxfun(@rdivide,feat,sum(feat,2));
        feat=sqrt(feat);
        contFeat(:,(1:atomN)+atomN*counter)=feat;
        counter=counter+1;
    end
end

function [score lh lw apprDist wParentFeat wChildFeat]=computeObj(parentInt,childInt,parentSW,childSW,parentStepsize,childStepsize,w,lambda,c_appr,c_proj,spry,childScore)

[parentApprFeat MP NP parentHMark parentWMark atomP]=getBatchFeat(spry,parentInt,parentSW,parentStepsize);
[childApprFeat MC NC childHMark childWMark atomC]=getBatchFeat(spry,childInt,childSW,childStepsize);
parentFeat=parentApprFeat(:,1:atomP);
childFeat=childApprFeat(:,1:atomC);

% MP=size(parentInt,1);
% NP=size(parentInt,2);
% MC=size(childInt,1);
% NC=size(childInt,2);
% parentInt=reshape(parentInt,[],size(parentInt,3));
% childInt=reshape(childInt,[],size(childInt,3));
% 
% ppp=reshape(1:MP*NP,[MP NP]);
% [ignore parentHMark parentWMark]=genSize([MP NP],parentSW,parentStepsize);
% hSwS=ppp(parentHMark,parentWMark);
% hSwS=hSwS(:);
% [uu vv]=ind2sub([MP NP],hSwS);
% uuEnd=uu+parentSW(1)-1;
% vvEnd=vv+parentSW(2)-1;
% hEwE=sub2ind([MP NP],uuEnd,vvEnd);
% hSwE=sub2ind([MP NP],uu,vvEnd);
% hEwS=sub2ind([MP NP],uuEnd,vv);
% parentFeat=parentInt(hEwE,:)+parentInt(hSwS,:)-parentInt(hSwE,:)-parentInt(hEwS,:);
% parentFeat=bsxfun(@rdivide,parentFeat,sum(parentFeat,2));
% parentFeat=sqrt(parentFeat);
% 
% ppp=reshape(1:MC*NC,[MC NC]);
% [ignore childHMark childWMark]=genSize([MC NC],childSW,childStepsize);
% hSwS=ppp(childHMark,childWMark);
% hSwS=hSwS(:);
% [uu vv]=ind2sub([MC NC],hSwS);
% uuEnd=uu+childSW(1)-1;
% vvEnd=vv+childSW(2)-1;
% hEwE=sub2ind([MC NC],uuEnd,vvEnd);
% hSwE=sub2ind([MC NC],uu,vvEnd);
% hEwS=sub2ind([MC NC],uuEnd,vv);
% childFeat=childInt(hEwE,:)+childInt(hSwS,:)-childInt(hSwE,:)-childInt(hEwS,:);
% childFeat=bsxfun(@rdivide,childFeat,sum(childFeat,2));
% childFeat=sqrt(childFeat);

% compute distance between window appearances
apprDist=pdist2(parentApprFeat,childApprFeat);
% compute distance on the projection line w
wParentFeat=(parentFeat*w)';
wChildFeat=(childFeat*w)';
projCost=max(lambda-bsxfun(@minus,wParentFeat',wChildFeat),0);
overallObj=-c_proj*projCost-c_appr*apprDist;
if exist('childScore','var')
    % child is not a leaf node
    overallObj=bsxfun(@plus,overallObj,childScore(:)');
end

[score maxI]=max(overallObj,[],2);
score=reshape(score,genSize([MP NP],parentSW,parentStepsize));
[lh lw]=ind2sub(genSize([MC NC],childSW,childStepsize),maxI);
lh=reshape(lh,genSize([MP NP],parentSW,parentStepsize));
lw=reshape(lw,genSize([MP NP],parentSW,parentStepsize));
wParentFeat=reshape(wParentFeat,[length(parentHMark) length(parentWMark)]);
wChildFeat=reshape(wChildFeat,[length(childHMark) length(childWMark)]);

apprIdx=sub2ind(size(apprDist),vec(1:size(apprDist,1)),maxI);
apprDist=reshape(apprDist(apprIdx),[length(parentHMark) length(parentWMark)]);


function stepsize=estimateStepsize(parentSW,parentInt,MAXWIN)
% assume the parentSW in decreasing order 
mapSize=cellfun(@(x)size(x),parentInt,'uniformoutput',false);
mapSize=cat(1,mapSize{:});
minParentSW=parentSW(end,:);
stepsize=zeros(size(mapSize,1),1);
for i = 1:size(mapSize,1)
    pixelN=prod(mapSize(i,1:2)-minParentSW+1);
    ratio=pixelN/MAXWIN;
    stepsize(i)=max(1,ceil(sqrt(ratio)));
end

function [size hmark wmark]=genSize(mapSize,sw,stepsize)
hmark=1:stepsize:mapSize(1)-sw(1)+1;
wmark=1:stepsize:mapSize(2)-sw(2)+1;
size(1)=length(hmark);
size(2)=length(wmark);



function [child parent]=passmsg(child,parent,w,lambda,c_appr,c_proj,spry,integralDir)
MAXWIN=2600;
childImgIdx=child.imgIdx;
parentImgIdx=parent.imgIdx;
parentSW=parent.sw;
childSW=child.sw;
if isfield(child,'score')
    isLeaf=false;
    childScore=child.score;
    childInt=child.cache.int;
    childScales=child.scales;
    childScaledSize=child.scaledSize;
    %child.cache=[];% delete cached info
    child=rmfield(child,'cache');
    childStepsize=child.stepsize;
    childPyramidSize=child.pyramidSize;
    childResp=child.resp;
else
    isLeaf=true;
    [childInt childScales childScaledSize]=loadIntegralMap(integralDir,childImgIdx);
    childStepsize=estimateStepsize(childSW,childInt,MAXWIN);
    childPyramidSize=cellfun(@(x)size(x),childInt,'uniformoutput',false);
    childPyramidSize=cat(1,childPyramidSize{:});
    childResp=cell(size(childSW,1),1);
    for i = 1:length(childResp)
        childResp{i}=cell(length(childInt),1);
    end
end


[parentInt parentScales parentScaledSize]=loadIntegralMap(integralDir,parentImgIdx);
parentStepsize=estimateStepsize(parentSW,parentInt,MAXWIN);

total_score=cell(size(parentSW,1),1);
total_lh=cell(size(parentSW,1),1);
total_lw=cell(size(parentSW,1),1);
total_Ik=cell(size(parentSW,1),1);
total_lad=cell(size(parentSW,1),1);
parentResp=cell(size(parentSW,1),1);

for p = 1:size(parentSW,1)
    total_score{p}=cell(length(parentInt),1);
    total_lh{p}=cell(length(parentInt),1);
    total_lw{p}=cell(length(parentInt),1);
    total_Ik{p}=cell(length(parentInt),1);
    total_lad{p}=cell(length(parentInt),1);
    parentResp{p}=cell(length(parentInt),1);
    for i = 1:length(parentInt)
        if size(parentInt{i},1)<parentSW(p,1) || size(parentInt{i},2)<parentSW(p,2)
            continue;
        end
        respSize=genSize(size(parentInt{i}),parentSW(p,:),parentStepsize(i));
        respH=respSize(1);
        respW=respSize(2);
        score=zeros(respH,respW,size(childSW,1)*length(childInt));
        lh=zeros(respH,respW,size(childSW,1)*length(childInt));
        lw=zeros(respH,respW,size(childSW,1)*length(childInt));
        lad=zeros(respH,respW,size(childSW,1)*length(childInt));
        layer0=reshape(1:respH*respW,[respH respW]);
        for q = 1:size(childSW,1)
            for j = 1:length(childInt)
                layerIdx=(q-1)*length(childInt)+j;
                if size(childInt{j},1)<childSW(q,1) || size(childInt{j},2)<childSW(q,2)
                    score(:,:,layerIdx)=-inf;
                    lh(:,:,layerIdx)=nan;
                    lw(:,:,layerIdx)=nan;
                else
                    if isLeaf
                        [score(:,:,layerIdx) lh(:,:,layerIdx) lw(:,:,layerIdx) lad(:,:,layerIdx) wParentFeat wChildFeat]=computeObj(parentInt{i},childInt{j},parentSW(p,:),childSW(q,:),parentStepsize(i),childStepsize(j),w,lambda,c_appr,c_proj,spry);
                    else
                        [score(:,:,layerIdx) lh(:,:,layerIdx) lw(:,:,layerIdx) lad(:,:,layerIdx) wParentFeat wChildFeat]=computeObj(parentInt{i},childInt{j},parentSW(p,:),childSW(q,:),parentStepsize(i),childStepsize(j),w,lambda,c_appr,c_proj,spry,childScore{q}{j});
                        % childPathIdx=sub2ind(size(childScore{q}{j}),vec(lh(:,:,layerIdx)),vec(lw(:,:,layerIdx)));
                        % childPathScore=reshape(childScore{q}{j}(childPathIdx),[size(score,1) size(score,2)]);
                        % score(:,:,layerIdx)=score(:,:,layerIdx)+childPathScore;
                    end
                    if isempty(parentResp{p}{i})
                        parentResp{p}{i}=wParentFeat;
                    end
                    if isempty(childResp{q}{j})
                        childResp{q}{j}=wChildFeat;
                    end
                end
            end
        end
        [score maxLayer]=max(score,[],3);
        idx=layer0+respH*respW*(maxLayer-1);
        lh=lh(idx);
        lw=lw(idx);
        lad=lad(idx);
        
        if isfield(parent,'score')
            score=score+parent.score{p}{i};
        end
        
        total_score{p}{i}=score;
        total_lh{p}{i}=lh;
        total_lw{p}{i}=lw;
        total_Ik{p}{i}=maxLayer;
        total_lad{p}{i}=lad;
    end
end
parentPyramidSize=cellfun(@(x)size(x),parentInt,'uniformoutput',false);
parentPyramidSize=cat(1,parentPyramidSize{:});

parent.score=total_score;
parent.stepsize=parentStepsize;
parent.pyramidSize=parentPyramidSize;
parent.scales=parentScales;
parent.scaledSize=parentScaledSize;
parent.resp=parentResp;
child.lh=total_lh;
child.lw=total_lw;
child.lk=total_Ik;
child.lad=total_lad;
child.pyramidSize=childPyramidSize;
child.stepsize=childStepsize;
child.scales=childScales;
child.scaledSize=childScaledSize;
child.resp=childResp;

% cache stuffs in memory thus avoid re-reading
parent.cache.int=parentInt;



% % DFS search along the structure
% stack=[];
% [curInt scales scaledSize]=loadIntegralMap(integralImgDir,getImgIdx(structure,structure.root));
% curMap=computeObjForSingleNode([],[],curInt,sw);
% stack=insertStack(stack,structure.root,curMap,curInt);
% while ~isempty(stack)
%     % Travel across nodes
%     [idx map int stack]=popStack(stack);
%     children=structure.data(idx).children;
%     
%     % Do things at this node
%     for i = length(children):-1:1% let's do it in an inverse way to make sure they pop out in the order matching "children" variable
%         [childInt scales scaledSize]=loadIntegralMap(integralImgDir,getImgIdx(structure,children(i)));
%         childMap=computeObjForSingleNode(int,map,childInt,sw);
%         stack=insertStack(stack,children(i),childMap,childInt);
%     end
% end
% 
% function imgIdx=getImgIdx(structure,nodeIdx)
% imgIdx=structure.data(nodeIdx).imgIdx;
% 
% function stack=insertStack(stack,idx,map,int)
% idx=flipud(idx);
% for i = 1:length(idx)
%     stack(end+1).nodeIdx=idx;
%     stack(end+1).map=map;
%     stack(end+1).int=int;
% end
% 
% function [idx map int stack]=popStack(stack)
% if isempty(stack)
%     idx=[];
%     map=[];
%     int=[];
%     return;
% end
% idx=stack(end).nodeIdx;
% map=stack(end).map;
% int=stack(end).int;
% stack(end)=[];
% 
% function childMap=computeObjForSingleNode(parentInt,parentMap,childInt,sw)
% if isempty(parentInt) || isempty(parentMap)
%     % for root who does not have parents
%     childMap=cell(size(sw,1),1);
%     for i = 1:size(sw,1)
%         childMap{i}=cell(length(childInt),1);
%         for j = 1:length(childInt)
%             if size(childInt{j},1)-sw(i,1)+1<=0 || size(childInt{j},2)-sw(i,2)+1<=0
%                 childMap{i}=childMap{i}(1:j-1);
%                 break;
%             end
%             childMap{i}{j}=zeros(size(childInt{j},1)-sw(i,1)+1,size(childInt{j},2)-sw(i,2)+1);
%         end
%     end
% else
%     % compute the interface between parent and child 
%     parentInt
% end







