function [resstruct,feat_pyramid] = esvm_detect_large(I, models, params)
% Localize a set of models in an image.
% function [resstruct,feat_pyramid] = esvm_detect(I, models, params)
%
% If there is a small number of models (such as in per-exemplar
% mining), then fconvblas is used for detection.  If the number is
% large, then the BLOCK feature matrix method (with a single matrix
% multiplication) is used.
%
% NOTE: These local detections can be pooled with esvm_pool_exemplars_dets.m
%
% I: Input image (or already precomputed pyramid)
% models: A cell array of models to localize inside this image
%   models{:}.model.w: Learned template
%   models{:}.model.b: Learned template's offset
% params: Localization parameters (see esvm_get_default_params.m)
%
% resstruct: Sliding window output struct with 
%   resstruct.bbs{:}: Detection boxes and pyramid locations
%   resstruct.xs{:}: Detection features
% feat_pyramid: The Feature pyramid output
%
% Copyright (C) 2011-12 by Tomasz Malisiewicz
% All rights reserved.
% 
% This file is part of the Exemplar-SVM library and is made
% available under the terms of the MIT license (see COPYING file).
% Project homepage: https://github.com/quantombone/exemplarsvm

if isempty(models)
  fprintf(1,'Warning: empty models in esvm_detect\n');
  resstruct.bbs{1} = zeros(0,0);
  resstruct.xs{1} = zeros(0,0);
  feat_pyramid = [];
  return;
end

if ~iscell(models)
  models = {models};
end

if isfield(models{1},'mining_params') && ~exist('params','var')
  params = models{1}.mining_params;
elseif ~exist('params','var')
  params = esvm_get_default_params;
end

if ~isfield(params,'nnmode')
 params.nnmode = '';
end

if ~isfield(params,'save_fired_HOG_loc')
    params.save_fired_HOG_loc = 0;
end

doflip = params.detect_add_flip;

params.detect_add_flip = 0;

[rs1, t1] = esvm_detectdriver(I, models, params);

[rs1 preserved_idx] = prune_nms(rs1, params);
if doflip == 1
  params.detect_add_flip = 1;
  [rs2, t2] = esvm_detectdriver(I, models, params);
  [rs2 preserved_idx] = prune_nms(rs2, params);
else %If there is no flip, then we are done
  resstruct = rs1;
  feat_pyramid = t1;
  return;
end

%If we got here, then the flip was turned on and we need to concatenate
%results
for q = 1:length(rs1.bbs)
  rs1.xs{q} = cat(2,rs1.xs{q}, rs2.xs{q});
  rs1.bbs{q} = cat(1,rs1.bbs{q},rs2.bbs{q});
end
resstruct = rs1;
%Concatenate normal and LR pyramids
feat_pyramid = cat(1,t1,t2);

function [resstruct,t] = esvm_detectdriver(I, models, ...
                                             params)
[resstruct,t] = esvm_detectdriverBLOCK(I, models, params);
for i = 1:length(resstruct.xs)
    for j = 1:size(resstruct.xs{i},2)
        xs{i}{1,j} = resstruct.xs{i}(:,j);
    end
end
try
    resstruct.xs = xs;
catch
end

function [resstruct,t] = esvm_detectdriverBLOCK(I, models, ...
                                             params)

%%HERE is the chunk version of exemplar localization

N = length(models);
ws = cellfun2(@(x)x.model.w,models);
bs = cellfun(@(x)x.model.b,models)';
bs = reshape(bs,[],1);
sizes1 = cellfun(@(x)x.model.hg_size(1),models);
sizes2 = cellfun(@(x)x.model.hg_size(2),models);

S = [max(sizes1(:)) max(sizes2(:))];
fsize = params.init_params.features();
templates = zeros(S(1),S(2),fsize,length(models));
templates_x = zeros(S(1),S(2),fsize,length(models));
template_masks = zeros(S(1),S(2),fsize,length(models));

for i = 1:length(models)
  t = zeros(S(1),S(2),fsize);
  t(1:models{i}.model.hg_size(1),1:models{i}.model.hg_size(2),:) = ...
      models{i}.model.w;

  templates(:,:,:,i) = t;
  template_masks(:,:,:,i) = repmat(double(sum(t.^2,3)>0),[1 1 fsize]);

  if (~isempty(params.nnmode)) || ...
        (isfield(params,'wtype') && ...
         strcmp(params.wtype,'dfun')==1)
    x = zeros(S(1),S(2),fsize);
    x(1:models{i}.model.hg_size(1),1:models{i}.model.hg_size(2),:) = ...
        reshape(models{i}.model.x(:,1),models{i}.model.hg_size);
    templates_x(:,:,:,i) = x;
  end
end

%maskmat = repmat(template_masks,[1 1 1 fsize]);
%maskmat = permute(maskmat,[1 2 4 3]);
%templates_x  = templates_x .* maskmat;

sbin = models{1}.model.init_params.sbin;

if isfield(params,'precomputation')
    if (params.detect_add_flip == 1)
        t = params.precomputation.flip;
    else
        t = params.precomputation.orig;
    end
else
    t = get_pyramid(I, sbin, params);
end


resstruct.padder = t.padder;
resstruct.r = -inf(N,1);
resstruct.bbs = cell(N,1);
resstruct.xs = cell(N,1);
resstruct.details.highest_response_info = [];
if params.save_fired_HOG_loc
    resstruct.details.HOG_loc = cell(N,1);
end

available_size = cellfun(@(x)[size(x,1) size(x,2)]-S+1,t.hog,'uniformoutput',false);
available_size = cat(1,available_size{:});
val_idx = find(sum(available_size<=0,2)==0);
pyr_N = prod(available_size(val_idx,:),2);
exemplar_matrix = reshape(templates,[],size(templates,4));

% set up default loading parameters
if ~isfield(params,'patch_loading')
    params.patch_loading = 30000;
end
if ~isfield(params,'model_loading')
    params.model_loading = 3500;
end


level_counter = 0;
while level_counter < length(pyr_N)
    patch_counter = 0;
    inner_level_counter = 0;
    while patch_counter < params.patch_loading && ...
            level_counter+inner_level_counter<length(pyr_N)
        patch_counter = patch_counter + pyr_N(level_counter+inner_level_counter+1);
        inner_level_counter = inner_level_counter + 1;
    end

    sumN = patch_counter;
    X = zeros(S(1)*S(2)*fsize,sumN);
    offsets = cell(inner_level_counter, 1);
    uus = cell(inner_level_counter,1);
    vvs = cell(inner_level_counter,1);
    counter = 1;
    for i = 1:inner_level_counter
      s = size(t.hog{val_idx(level_counter+i)});
      NW = s(1)*s(2);
      ppp = reshape(1:NW,s(1),s(2));
      curf = reshape(t.hog{val_idx(level_counter+i)},[],fsize);
      b = im2col(ppp,[S(1) S(2)]);
      offsets{i} = b(1,:);
      offsets{i}(end+1,:) = val_idx(level_counter+i);
      for j = 1:size(b,2)
       X(:,counter) = reshape(curf(b(:,j),:),[],1);
       counter = counter + 1;
      end
      [uus{i},vvs{i}] = ind2sub(s,offsets{i}(1,:));
    end
    if isempty(X)
        return;
    end
    offsets = cat(2,offsets{:});
    uus = cat(2,uus{:});
    vvs = cat(2,vvs{:});
    modelBatchLoading = unSyncDistribute(N,params.model_loading);
    for modelBatchNumerator= 1:length(modelBatchLoading)
        modelBatchIdx = modelBatchLoading(modelBatchNumerator);
        modelValIdx = unSyncMapper(modelBatchIdx,N,params.model_loading);
        if isempty(params.nnmode)
          %nnmode 0: Apply linear classifiers by performing one large matrix
          %multiplication and subtract bias
            r = exemplar_matrix(:,modelValIdx)' * X;
            r = bsxfun(@minus, r, bs(modelValIdx));
        elseif strcmp(params.nnmode,'elda') == 1
            r = exemplar_matrix(:,modelValIdx)' * X;
            r = bsxfun(@minus, r, bs(modelValIdx));
            [threshScore ignore] = psort(-r(:),numel(r)*0.0005);
            params.detect_keep_threshold = -threshScore(end);
        else
          error('invalid nnmode=%s\n',params.nnmode);
        end
        
        for exid = 1:length(modelValIdx)
          goods = find(r(exid,:) >= params.detect_keep_threshold);
          if isempty(goods)
            continue    
          end
          modelIdx = modelValIdx(exid);
          [sorted_scores,bb] = ...
              psort(-r(exid,goods)',...
                    min(params.detect_max_windows_per_exemplar, ...
                        length(goods)));
          bb = goods(bb);
          sorted_scores = -sorted_scores';
          resstruct.xs{modelIdx} = [resstruct.xs{modelIdx} X(:,bb)];
          levels = offsets(2,bb);
          scales = t.scales(levels);
          curuus = uus(bb);
          curvvs = vvs(bb);
          o = [curuus' curvvs'] - t.padder;
          bbs = ([o(:,2) o(:,1) o(:,2)+size(ws{modelIdx},2) ...
                   o(:,1)+size(ws{modelIdx},1)] - 1) .* ...
                     repmat(sbin./scales',1,4) + 1 + repmat([0 0 -1 ...
                            -1],length(scales),1);
          bbs(:,5:12) = 0;
          bbs(:,5) = (1:size(bbs,1));
          bbs(:,6) = modelIdx;
          bbs(:,8) = scales;
          bbs(:,9) = uus(bb);
          bbs(:,10) = vvs(bb);
          bbs(:,12) = sorted_scores;
          if (params.detect_add_flip == 1)
            bbs = flip_box(bbs,t.size);
            bbs(:,7) = 1;
          end
          resstruct.bbs{modelIdx} = [resstruct.bbs{modelIdx};bbs];
        end
    end
    level_counter = level_counter + inner_level_counter;
end


if params.detect_save_features == 0
  resstruct.xs = cell(N,1);
end
%fprintf(1,'\n');

function [rs preserved_idx] = prune_nms(rs, params)
%Prune via nms to eliminate redundant detections

%If the field is missing, or it is set to 1, then we don't need to
%process anything.  If it is zero, we also don't do NMS.
preserved_idx = [];
if ~isfield(params,'detect_exemplar_nms_os_threshold') || (params.detect_exemplar_nms_os_threshold >= 1) ...
      || (params.detect_exemplar_nms_os_threshold == 0)
  return;
end

[rs.bbs preserved_idx] = cellfun(@(x)esvm_nms(x,params.detect_exemplar_nms_os_threshold),...
    rs.bbs,'uniformoutput',false);

if ~isempty(rs.xs)
  for i = 1:length(rs.bbs)
    if ~isempty(rs.xs{i})
      %NOTE: the fifth field must contain elements
      rs.xs{i} = rs.xs{i}(:,rs.bbs{i}(:,5) );
    end
  end
end

function t = get_pyramid(I, sbin, params)
%Extract feature pyramid from variable I (which could be either an image,
%or already a feature pyramid)

if isnumeric(I)
  if (params.detect_add_flip == 1)
    I = flip_image(I);
  else    
    %take unadulterated "aka" un-flipped image
  end
  
  clear t
  t.size = size(I);

  %Compute pyramid
  [t.hog, t.scales] = esvm_pyramid(I, params);
  t.padder = params.detect_pyramid_padding;
  for level = 1:length(t.hog)
    t.hog{level} = padarray(t.hog{level}, [t.padder t.padder 0], 0);
  end
  
  minsizes = cellfun(@(x)min([size(x,1) size(x,2)]), t.hog);
  t.hog = t.hog(minsizes >= t.padder*2);
  t.scales = t.scales(minsizes >= t.padder*2);  
else
  fprintf(1,'Already found features\n');
  
  if iscell(I)
    if params.detect_add_flip==1
      t = I{2};
    else
      t = I{1};
    end
  else
    t = I;
  end
end

