function [resstruct,feat_pyramid] = esvm_postTopK_detect(I, models, postTopK, params)
% Localize a set of models in an image.
% function [resstruct,feat_pyramid] = esvm_detect(I, models, params)
%
% If there is a small number of models (such as in per-exemplar
% mining), then fconvblas is used for detection.  If the number is
% large, then the BLOCK feature matrix method (with a single matrix
% multiplication) is used.
%
% NOTE: These local detections can be pooled with esvm_pool_exemplars_dets.m
%
% I: Input image (or already precomputed pyramid)
% models: A cell array of models to localize inside this image
%   models{:}.model.w: Learned template
%   models{:}.model.b: Learned template's offset
% params: Localization parameters (see esvm_get_default_params.m)
%
% resstruct: Sliding window output struct with 
%   resstruct.bbs{:}: Detection boxes and pyramid locations
%   resstruct.xs{:}: Detection features
% feat_pyramid: The Feature pyramid output
%
% Copyright (C) 2011-12 by Tomasz Malisiewicz
% All rights reserved.
% 
% This file is part of the Exemplar-SVM library and is made
% available under the terms of the MIT license (see COPYING file).
% Project homepage: https://github.com/quantombone/exemplarsvm

if isempty(models)
  fprintf(1,'Warning: empty models in esvm_detect\n');
  resstruct.bbs{1} = zeros(0,0);
  resstruct.xs{1} = zeros(0,0);
  feat_pyramid = [];
  return;
end

if ~iscell(models)
  models = {models};
end

if isfield(models{1},'mining_params') && ~exist('params','var')
  params = models{1}.mining_params;
elseif ~exist('params','var')
  params = esvm_get_default_params;
end

if ~isfield(params,'nnmode')
 params.nnmode = '';
end

if ~isfield(params,'save_fired_HOG_loc')
    params.save_fired_HOG_loc = 0;
end

doflip = params.detect_add_flip;

params.detect_add_flip = 0;

[rs1, t1] = esvm_detectdriverBLOCK(I, models, postTopK, params);

[rs1 preserved_idx] = prune_nms(rs1, params);
% prune HOG loc entry if necessary
if isfield(rs1.details,'HOG_loc') 
    rs1.details.HOG_loc = cellfun(@(x,y)x(y,:),rs1.details.HOG_loc,preserved_idx,...
        'uniformoutput',false);
end

if doflip == 1
  params.detect_add_flip = 1;
  [rs2, t2] = esvm_detectdriverBLOCK(I, models, postTopK, params);
  [rs2 preserved_idx] = prune_nms(rs2, params);
  % prune HOG loc entry if necessary
  if isfield(rs2.details,'HOG_loc')
    rs2.details.HOG_loc = cellfun(@(x,y)x(y,:),rs2.details.HOG_loc,preserved_idx,...
        'uniformoutput',false);
  end

else %If there is no flip, then we are done
  resstruct = rs1;
  feat_pyramid = t1;
  return;
end

%If we got here, then the flip was turned on and we need to concatenate
%results
if params.save_fired_HOG_loc
    for q = 1:length(rs1.bbs)
      rs1.xs{q} = cat(2,rs1.xs{q}, rs2.xs{q});
      rs1.bbs{q} = cat(1,rs1.bbs{q},rs2.bbs{q});
      rs1.details.HOG_loc{q} = cat(1,rs1.details.HOG_loc{q},rs2.details.HOG_loc{q});
    end
else
    for q = 1:length(rs1.bbs)
      rs1.xs{q} = cat(2,rs1.xs{q}, rs2.xs{q});
      rs1.bbs{q} = cat(1,rs1.bbs{q},rs2.bbs{q});
    end
end

if isfield(rs2.details,'highest_response_info')
    rs1.details.highest_response_info_flip = rs2.details.highest_response_info;
end
if isfield(rs1,'r') && isfield(rs2,'r')
    rs1.r = [rs1.r;rs2.r];
end

if isfield(rs2.details,'responseMap')
    rs1.details.responseMap = [rs1.details.responseMap; rs2.details.responseMap];
end
resstruct = rs1;

%Concatenate normal and LR pyramids
feat_pyramid = cat(1,t1,t2);


function [resstruct,t] = esvm_detectdriverBLOCK(I, models, ...
                                             postTopK,params)

%%HERE is the chunk version of exemplar localization

N = length(models);
ws = cellfun2(@(x)x.model.w,models);
bs = cellfun(@(x)x.model.b,models)';
bs = reshape(bs,[],1);
sizes1 = cellfun(@(x)x.model.hg_size(1),models);
sizes2 = cellfun(@(x)x.model.hg_size(2),models);

S = [max(sizes1(:)) max(sizes2(:))];
fsize = params.init_params.features();
templates = zeros(S(1),S(2),fsize,length(models));
templates_x = zeros(S(1),S(2),fsize,length(models));
template_masks = zeros(S(1),S(2),fsize,length(models));

for i = 1:length(models)
  t = zeros(S(1),S(2),fsize);
  t(1:models{i}.model.hg_size(1),1:models{i}.model.hg_size(2),:) = ...
      models{i}.model.w;

  templates(:,:,:,i) = t;
  template_masks(:,:,:,i) = repmat(double(sum(t.^2,3)>0),[1 1 fsize]);

  if (~isempty(params.nnmode)) || ...
        (isfield(params,'wtype') && ...
         strcmp(params.wtype,'dfun')==1)
    x = zeros(S(1),S(2),fsize);
    x(1:models{i}.model.hg_size(1),1:models{i}.model.hg_size(2),:) = ...
        reshape(models{i}.model.x(:,1),models{i}.model.hg_size);
    templates_x(:,:,:,i) = x;
  end
end

%maskmat = repmat(template_masks,[1 1 1 fsize]);
%maskmat = permute(maskmat,[1 2 4 3]);
%templates_x  = templates_x .* maskmat;

sbin = models{1}.model.init_params.sbin;

if isfield(params,'precomputation')
    if (params.detect_add_flip == 1)
        t = params.precomputation.flip;
    else
        t = params.precomputation.orig;
    end
else
    t = get_pyramid(I, sbin, params);
end


resstruct.padder = t.padder;

available_size = cellfun(@(x)[size(x,1) size(x,2)]-S+1,t.hog,'uniformoutput',false);
available_size = cat(1,available_size{:});
val_idx = find(sum(available_size<=0,2)==0);

pyr_N = prod(available_size(val_idx,:),2);
sumN = sum(pyr_N);

X = zeros(S(1)*S(2)*fsize,sumN);
offsets = cell(length(val_idx), 1);
uus = cell(length(val_idx),1);
vvs = cell(length(val_idx),1);

counter = 1;
for i = 1:length(val_idx)
  s = size(t.hog{val_idx(i)});
  NW = s(1)*s(2);
  ppp = reshape(1:NW,s(1),s(2));
  curf = reshape(t.hog{val_idx(i)},[],fsize);
  b = im2col(ppp,[S(1) S(2)]);

  offsets{i} = b(1,:);
  offsets{i}(end+1,:) = val_idx(i);
  
%   X_i = arrayfun(@(x)reshape(curf(b(:,x),:),[],1),[1:size(b,2)],'UniformOutput',false);
%   X(:,counter:size(b,2)+counter-1) = cat(2,X_i{:});
%   counter = counter + size(b,2);
  
  for j = 1:size(b,2)
   X(:,counter) = reshape(curf(b(:,j),:),[],1);
   counter = counter + 1;
  end
  
  [uus{i},vvs{i}] = ind2sub(s,offsets{i}(1,:));
end

offsets = cat(2,offsets{:});

uus = cat(2,uus{:});
vvs = cat(2,vvs{:});

% m.model.w = zeros(S(1),S(2),fsize);
% m.model.b = 0;
% temp_params = params;
% temp_params.detect_save_features = 1;
% temp_params.detect_exemplar_nms_os_threshold = 1.0;
% temp_params.max_models_before_block_method = 1;
% temp_params.detect_max_windows_per_exemplar = 28000;

% [rs] = esvm_detect(I, {m}, temp_params);
% X2=cat(2,rs.xs{1}{:});
% bbs2 = rs.bbs{1};

exemplar_matrix = reshape(templates,[],size(templates,4));
% pick the postTopK filters at each window
completeR = exemplar_matrix' * X;
completeR = bsxfun(@minus, completeR, bs);
[sortedCompleteRVal sortedCompleteRIdx] = sort(completeR,1,'descend');
postIdx = sortedCompleteRIdx(1:postTopK,:);
r = -inf(size(completeR));
postIdx = bsxfun(@plus, postIdx, [0:1:(size(completeR,2)-1)]*size(completeR,1));
postIdx = postIdx(:);
r(postIdx) = completeR(postIdx);

resstruct.r = max(r,[],2);
% resstruct.r_avg = mean(r,2);
resstruct.bbs = cell(N,1);
resstruct.xs = cell(N,1);
resstruct.details.highest_response_info = [];
if params.save_fired_HOG_loc
    resstruct.details.HOG_loc = cell(N,1);
end

if isempty(X)
    return;
end

if params.getResponseMap
    % collect some data
    resstruct.details.responseMap = single(r);
    resstruct.details.hog_levels = offsets(2,:);
    resstruct.details.hog_scales = t.scales;
    resstruct.details.hog_padder = t.padder;
    resstruct.details.uus = uus;
    resstruct.details.vvs = vvs;
    resstruct.details.hog_size = cellfun(@(x)size(x),t.hog,'UniformOutput',false);
    resstruct.details.hog_size = cat(1,resstruct.details.hog_size{:});
end

for exid = 1:N

  goods = find(r(exid,:) >= params.detect_keep_threshold);
  
  if isempty(goods)
    % collect info about highest response
    [ignore bb_max]= max(r(exid,:));
    hi_level = offsets(2,bb_max);
    hi_uus = uus(bb_max);
    hi_vvs = vvs(bb_max);
    resstruct.details.highest_response_info(exid,:) = [hi_uus hi_vvs hi_level params.detect_add_flip];
    continue    
  end
  
  [sorted_scores,bb] = ...
      psort(-r(exid,goods)',...
            min(params.detect_max_windows_per_exemplar, ...
                length(goods)));
  bb = goods(bb);

  sorted_scores = -sorted_scores';

  resstruct.xs{exid} = X(:,bb);
  
  levels = offsets(2,bb);
  scales = t.scales(levels);
  curuus = uus(bb);
  curvvs = vvs(bb);
  
  if params.save_fired_HOG_loc
    resstruct.details.HOG_loc{exid} = [curuus' curvvs' levels' params.detect_add_flip*ones(length(curuus),1)];
  end
  
  % record info about highest response
  resstruct.details.highest_response_info(exid,:) = [curuus(1) curvvs(1) levels(1) params.detect_add_flip];
  
  o = [curuus' curvvs'] - t.padder;

  bbs = ([o(:,2) o(:,1) o(:,2)+size(ws{exid},2) ...
           o(:,1)+size(ws{exid},1)] - 1) .* ...
             repmat(sbin./scales',1,4) + 1 + repmat([0 0 -1 ...
                    -1],length(scales),1);
  
  bbs(:,5:12) = 0;
  bbs(:,5) = (1:size(bbs,1));
  bbs(:,6) = exid;
  bbs(:,8) = scales;
  bbs(:,9) = uus(bb);
  bbs(:,10) = vvs(bb);
  bbs(:,12) = sorted_scores;
  
  if (params.detect_add_flip == 1)
    bbs = flip_box(bbs,t.size);
    bbs(:,7) = 1;
  end
  
  resstruct.bbs{exid} = bbs;
end

if params.detect_save_features == 0
  resstruct.xs = cell(N,1);
end
%fprintf(1,'\n');

function [rs preserved_idx] = prune_nms(rs, params)
%Prune via nms to eliminate redundant detections

%If the field is missing, or it is set to 1, then we don't need to
%process anything.  If it is zero, we also don't do NMS.
preserved_idx = [];
if ~isfield(params,'detect_exemplar_nms_os_threshold') || (params.detect_exemplar_nms_os_threshold >= 1) ...
      || (params.detect_exemplar_nms_os_threshold == 0)
  return;
end

[rs.bbs preserved_idx] = cellfun(@(x)esvm_nms(x,params.detect_exemplar_nms_os_threshold),...
    rs.bbs,'uniformoutput',false);

if ~isempty(rs.xs)
  for i = 1:length(rs.bbs)
    if ~isempty(rs.xs{i})
      %NOTE: the fifth field must contain elements
      rs.xs{i} = rs.xs{i}(:,rs.bbs{i}(:,5) );
    end
  end
end

function t = get_pyramid(I, sbin, params)
%Extract feature pyramid from variable I (which could be either an image,
%or already a feature pyramid)

if isnumeric(I)
  if (params.detect_add_flip == 1)
    I = flip_image(I);
  else    
    %take unadulterated "aka" un-flipped image
  end
  
  clear t
  t.size = size(I);

  %Compute pyramid
  [t.hog, t.scales] = esvm_pyramid(I, params);
  t.padder = params.detect_pyramid_padding;
  for level = 1:length(t.hog)
    t.hog{level} = padarray(t.hog{level}, [t.padder t.padder 0], 0);
  end
  
  minsizes = cellfun(@(x)min([size(x,1) size(x,2)]), t.hog);
  t.hog = t.hog(minsizes >= t.padder*2);
  t.scales = t.scales(minsizes >= t.padder*2);  
else
  fprintf(1,'Already found features\n');
  
  if iscell(I)
    if params.detect_add_flip==1
      t = I{2};
    else
      t = I{1};
    end
  else
    t = I;
  end
end

