function [curfeats template_h template_w] = get_padded_patch(I,bbox,init_params)
% Initialize the model representing patches. Therefore the key difference
% between this and the standard initialization procedure is the size of the
% model is determined by the average size of the common patches.
% Fanyi Xiao


if ~exist('init_params','var')
  init_params.sbin = 8;
  init_params.hg_size = [8 8];
  init_params.MAXDIM = 10;
end

if ~isfield(init_params,'MAXDIM')
  init_params.MAXDIM = 10;
  fprintf(1,'Default MAXDIM is %d\n',init_params.MAXDIM);
end

epsilon = 1e-8;

template_length = init_params.template_length;
im_w = size(I,2);
im_h = size(I,1);
bbox_w = bbox(3) - bbox(1) + 1;
bbox_h = bbox(4) - bbox(2) + 1;
wh_ratio = bbox_w/bbox_h;
% compute template height and width
if wh_ratio>=1
    template_w = template_length;
    template_h = max(1,round(template_w/wh_ratio));
else
    template_h = template_length;
    template_w = max(1,round(wh_ratio*template_h));
end

rescale_w = (init_params.sbin*template_w)/bbox_w;
rescale_h = (init_params.sbin*template_h)/bbox_h;
new_im_w = round(rescale_w*im_w);
new_im_h = round(rescale_h*im_h);
bbox([1 3]) = round(bbox([1 3])*rescale_w);
bbox([2 4]) = round(bbox([2 4])*rescale_h);
bbox = clip_to_image(bbox,[1 1 new_im_w new_im_h]);
cur_I = imresize(I,[new_im_h,new_im_w]);

%Create a blank image with the exemplar inside
Ibox = zeros(size(cur_I,1), size(cur_I,2));    
Ibox(bbox(2):bbox(4), bbox(1):bbox(3)) = 1;

%% NOTE: why was I padding this at some point and now I'm not???
%% ANSWER: doing the pad will create artifical gradients
ARTPAD = 0;
I_real_pad = pad_image(cur_I, ARTPAD);

%Get the hog feature pyramid for the entire image
clear params;
params.detect_levels_per_octave = 10;
params.init_params = init_params;
[f_real,scales] = esvm_pyramid(I_real_pad, params);

%Extract the regions most overlapping with Ibox from each level in the pyramid
[masker,sizer] = get_matching_masks(f_real, Ibox);

%Now choose the mask which is closest to N cells
[targetlvl, mask] = get_ncell_mask([template_h template_w],masker,sizer);
[uu,vv] = find(mask);
min_uu = min(uu);
max_uu = max(uu);
min_vv = min(vv);
max_vv = max(vv);

curfeats = pad_with_content(...
    f_real{targetlvl},min_uu,max_uu,min_vv,max_vv,template_length);
% curfeats = curfeats - mean(curfeats(:));

debug = 0;
if debug
    dispI = plot_bb(cur_I,bbox,'r');
    dispHOG = HOGpicture(curfeats);
    subplot_tight(1,2,1);
    imshow(dispI);
    title('Image');
    subplot_tight(1,2,2);
    imshow(dispHOG);
    title('HOG');
end


function [targetlvl,mask] = get_ncell_mask(temp_size, masker,sizer)
%Get a the mask and features, where mask is closest to NCELL cells
%as possible
diff = sum(abs(repmat(temp_size,[size(sizer,1) 1])-sizer),2);
[sortval sortind] = sort(diff,'ascend');

autoadjustsize = 0;
if autoadjustsize == 1
    [uu,vv] = find(masker{sortind(1)});
    mask = masker{sortind(1)}*0;
    targetlvl = sortind(1);
    min_y = min(uu);
    max_y = max(uu);
    min_x = min(vv);
    max_x = max(vv);
    mask(min_y:max_y,min_x:max_x) = 1;
else               
    hog_w = size(masker{sortind(1)},2);
    hog_h = size(masker{sortind(1)},1);
    [uu,vv] = find(masker{sortind(1)});
    min_y = min(uu);
    max_y = max(uu);
    min_x = min(vv);
    max_x = max(vv);

    if (min_y + temp_size(1) - 1) <= hog_h
        max_y = min_y + temp_size(1) - 1;
    elseif (max_y - temp_size(1) + 1) >= 1
        min_y = (max_y - temp_size(1) + 1);
    else
        max_y = hog_h;
    end

    if (min_x + temp_size(2) - 1) <= hog_w
        max_x = min_x + temp_size(2) - 1;

    elseif (max_x - temp_size(2) + 1) >= 1
        min_x = (max_x - temp_size(2) + 1);
    else
        max_x = hog_w;
    end
    mask = masker{sortind(1)}*0;
    targetlvl = sortind(1);
    mask(min_y:max_y,min_x:max_x) = 1;
    return;
end

function [masker,sizer] = get_matching_masks(f_real, Ibox)
%Given a feature pyramid, and a segmentation mask inside Ibox, find
%the best matching region per level in the feature pyramid

masker = cell(length(f_real),1);
sizer = zeros(length(f_real),2);

for a = 1:length(f_real)
  goods = double(sum(f_real{a}.^2,3)>0);
  
  masker{a} = max(0.0,min(1.0,imresize(Ibox,[size(f_real{a},1) size(f_real{a}, ...
                                                  2)])));
  [tmpval,ind] = max(masker{a}(:));
  masker{a} = (masker{a}>.1) & goods;

  if sum(masker{a}(:))==0
    [aa,bb] = ind2sub(size(masker{a}),ind);
    masker{a}(aa,bb) = 1;
  end
  [uu,vv] = find(masker{a});
  masker{a}(min(uu):max(uu),min(vv):max(vv))=1;
  sizer(a,:) = [range(uu)+1 range(vv)+1];
end


function curfeats = pad_with_content(hog,min_uu,max_uu,min_vv,max_vv,template_length)
h_pad = template_length - (max_uu-min_uu+1);
w_pad = template_length - (max_vv-min_vv+1);
hog_h = size(hog,1);
hog_w = size(hog,2);

if max_uu+h_pad<=hog_h
    forward_uu = max_uu+h_pad;
    h_recursive_pad_pixels = 0;
else
    h_recursive_pad_pixels = h_pad - (hog_h - max_uu);
    forward_uu = hog_h;
end
if max_vv+w_pad<=hog_w
    forward_vv = max_vv+w_pad;
    w_recursive_pad_pixels = 0;
else
    forward_vv = hog_w;
    w_recursive_pad_pixels = w_pad - (hog_w - max_vv);
end
curfeats = hog(min_uu:forward_uu,min_vv:forward_vv,:);
curfeats = padarray(curfeats,[h_recursive_pad_pixels w_recursive_pad_pixels],'post');






